@@ -2,16 +2,14 @@
 
 use std::{convert::Infallible, time::Duration};
 
-use async_trait::async_trait;
 use tokio::time;
 
-use deadpool::managed::{self, Object, PoolError, RecycleResult};
+use deadpool::managed::{self, Metrics, Object, PoolError, RecycleResult, Timeouts};
 
 type Pool = managed::Pool<Manager>;
 
 struct Manager {}
 
-#[async_trait]
 impl managed::Manager for Manager {
     type Type = usize;
     type Error = Infallible;
@@ -20,7 +18,7 @@ impl managed::Manager for Manager {
         Ok(0)
     }
 
-    async fn recycle(&self, _conn: &mut usize) -> RecycleResult<Infallible> {
+    async fn recycle(&self, _conn: &mut usize, _: &Metrics) -> RecycleResult<Infallible> {
         Ok(())
     }
 }
@@ -33,36 +31,43 @@ async fn basic() {
     let status = pool.status();
     assert_eq!(status.size, 0);
     assert_eq!(status.available, 0);
+    assert_eq!(status.waiting, 0);
 
     let obj0 = pool.get().await.unwrap();
     let status = pool.status();
     assert_eq!(status.size, 1);
     assert_eq!(status.available, 0);
+    assert_eq!(status.waiting, 0);
 
     let obj1 = pool.get().await.unwrap();
     let status = pool.status();
     assert_eq!(status.size, 2);
     assert_eq!(status.available, 0);
+    assert_eq!(status.waiting, 0);
 
     let obj2 = pool.get().await.unwrap();
     let status = pool.status();
     assert_eq!(status.size, 3);
     assert_eq!(status.available, 0);
+    assert_eq!(status.waiting, 0);
 
     drop(obj0);
     let status = pool.status();
     assert_eq!(status.size, 3);
     assert_eq!(status.available, 1);
+    assert_eq!(status.waiting, 0);
 
     drop(obj1);
     let status = pool.status();
     assert_eq!(status.size, 3);
     assert_eq!(status.available, 2);
+    assert_eq!(status.waiting, 0);
 
     drop(obj2);
     let status = pool.status();
     assert_eq!(status.size, 3);
     assert_eq!(status.available, 3);
+    assert_eq!(status.waiting, 0);
 }
 
 #[tokio::test]
@@ -78,19 +83,29 @@ async fn closing() {
     };
 
     tokio::task::yield_now().await;
-    assert_eq!(pool.status().available, -1);
+    assert_eq!(pool.status().available, 0);
+    assert_eq!(pool.status().waiting, 1);
 
     pool.close();
     tokio::task::yield_now().await;
     assert_eq!(pool.status().available, 0);
+    assert_eq!(pool.status().waiting, 0);
 
     assert!(matches!(join_handle.await.unwrap(), Err(PoolError::Closed)));
     assert!(matches!(pool.get().await, Err(PoolError::Closed)));
-    assert!(matches!(pool.try_get().await, Err(PoolError::Closed)));
+    assert!(matches!(
+        pool.timeout_get(&Timeouts {
+            wait: Some(Duration::ZERO),
+            ..pool.timeouts()
+        })
+        .await,
+        Err(PoolError::Closed)
+    ));
 
     drop(obj);
     tokio::task::yield_now().await;
     assert_eq!(pool.status().available, 0);
+    assert_eq!(pool.status().waiting, 0);
 }
 
 #[tokio::test]
@@ -129,6 +144,7 @@ async fn concurrent() {
     let status = pool.status();
     assert_eq!(status.size, 3);
     assert_eq!(status.available, 3);
+    assert_eq!(status.waiting, 0);
 
     let values = [
         pool.get().await.unwrap(),
@@ -149,13 +165,15 @@ async fn object_take() {
     let status = pool.status();
     assert_eq!(status.size, 2);
     assert_eq!(status.available, 0);
+    assert_eq!(status.waiting, 0);
 
-    drop(Object::take(obj0));
+    let _ = Object::take(obj0);
     let status = pool.status();
     assert_eq!(status.size, 1);
     assert_eq!(status.available, 0);
+    assert_eq!(status.waiting, 0);
 
-    drop(Object::take(obj1));
+    let _ = Object::take(obj1);
     let status = pool.status();
     assert_eq!(status.size, 0);
     assert_eq!(status.available, 0);
@@ -165,12 +183,14 @@ async fn object_take() {
     let status = pool.status();
     assert_eq!(status.size, 2);
     assert_eq!(status.available, 0);
+    assert_eq!(status.waiting, 0);
 
     drop(obj0);
     drop(obj1);
     let status = pool.status();
     assert_eq!(status.size, 2);
     assert_eq!(status.available, 2);
+    assert_eq!(status.waiting, 0);
 }
 
 #[tokio::test]
@@ -235,21 +255,45 @@ async fn resize_pool_grow_concurrent() {
     tokio::task::yield_now().await;
     assert_eq!(pool.status().max_size, 0);
     assert_eq!(pool.status().size, 0);
-    assert_eq!(pool.status().available, -1);
+    assert_eq!(pool.status().available, 0);
+    assert_eq!(pool.status().waiting, 1);
     pool.resize(1);
     assert_eq!(pool.status().max_size, 1);
     assert_eq!(pool.status().size, 0);
-    assert_eq!(pool.status().available, -1);
+    assert_eq!(pool.status().available, 0);
+    assert_eq!(pool.status().waiting, 1);
     tokio::task::yield_now().await;
     assert_eq!(pool.status().max_size, 1);
     assert_eq!(pool.status().size, 1);
     assert_eq!(pool.status().available, 0);
+    assert_eq!(pool.status().waiting, 0);
     let obj0 = join_handle.await.unwrap().unwrap();
     assert_eq!(pool.status().max_size, 1);
     assert_eq!(pool.status().size, 1);
     assert_eq!(pool.status().available, 0);
+    assert_eq!(pool.status().waiting, 0);
     drop(obj0);
     assert_eq!(pool.status().max_size, 1);
     assert_eq!(pool.status().size, 1);
     assert_eq!(pool.status().available, 1);
+    assert_eq!(pool.status().waiting, 0);
+}
+
+#[tokio::test]
+async fn retain() {
+    let mgr = Manager {};
+    let pool = Pool::builder(mgr).max_size(4).build().unwrap();
+    {
+        let _a = pool.get().await;
+        let _b = pool.get().await;
+        tokio::time::sleep(Duration::from_millis(5)).await;
+        let _c = pool.get().await;
+        tokio::time::sleep(Duration::from_millis(5)).await;
+    }
+    assert_eq!(pool.status().size, 3);
+    pool.retain(|_, metrics| metrics.age() <= Duration::from_millis(10));
+    assert_eq!(pool.status().size, 1);
+    tokio::time::sleep(Duration::from_millis(5)).await;
+    pool.retain(|_, metrics| metrics.age() <= Duration::from_millis(10));
+    assert_eq!(pool.status().size, 0);
 }
